/*
 * ProviewR   Open Source Process Control.
 * Copyright (C) 2005-2019 SSAB EMEA AB.
 *
 * This file is part of ProviewR.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ProviewR. If not, see <http://www.gnu.org/licenses/>
 *
 * Linking ProviewR statically or dynamically with other modules is
 * making a combined work based on ProviewR. Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * In addition, as a special exception, the copyright holders of
 * ProviewR give you permission to, from the build function in the
 * ProviewR Configurator, combine ProviewR with modules generated by the
 * ProviewR PLC Editor to a PLC program, regardless of the license
 * terms of these modules. You may copy and distribute the resulting
 * combined work under the terms of your choice, provided that every
 * copy of the combined work is accompanied by a complete copy of
 * the source code of ProviewR (the version used to produce the
 * combined work), being distributed under the terms of the GNU
 * General Public License plus this exception.
 */

#include <stdlib.h>
#include <string.h>

#include <iostream>

#include "glow_growconpoint.h"
#include "glow_draw.h"
#include "glow_exportflow.h"

GrowConPoint::GrowConPoint(GrowCtx* glow_ctx, const char* name, double x,
    double y, int cp_num, glow_eDirection d, int nodraw)
    : GlowConPoint(glow_ctx, x, y, cp_num, d), hot(0), pzero(ctx), highlight(0),
      inverse(0),
      arc(ctx, x - ctx->draw_delta, y - ctx->draw_delta, x + ctx->draw_delta,
          y + ctx->draw_delta, 0, 360, glow_eDrawType_Line, 2),
      user_data(NULL)
{
  strcpy(n_name, name);
  pzero.nav_zoom();

  if (ctx->grid_on) {
    double x_grid, y_grid;

    ctx->find_grid(p.x, p.y, &x_grid, &y_grid);
    p.posit(x_grid, y_grid);
  }
  arc.ll.posit(-ctx->draw_delta, -ctx->draw_delta);
  arc.ur.posit(ctx->draw_delta, ctx->draw_delta);
  if (!nodraw)
    ctx->set_dirty();
  get_node_borders();
}

GrowConPoint::~GrowConPoint()
{
  ctx->object_deleted(this);

  ctx->set_dirty();
  if (hot)
    ctx->gdraw->set_cursor(ctx->mw.window, glow_eDrawCursor_Normal);
}

void GrowConPoint::move(double delta_x, double delta_y, int grid)
{
  if (grid) {
    double x, y, x_grid, y_grid;

    /* Move to closest grid point */
    x = (x_right + x_left) / 2;
    y = (y_high + y_low) / 2;
    ctx->find_grid(x + delta_x / ctx->mw.zoom_factor_x,
        y + delta_y / ctx->mw.zoom_factor_y, &x_grid, &y_grid);
    trf.move(x_grid - x, y_grid - y);
    get_node_borders();
  } else {
    double dx = delta_x / ctx->mw.zoom_factor_x;
    double dy = delta_y / ctx->mw.zoom_factor_y;
    trf.move(dx, dy);
    x_right += dx;
    x_left += dx;
    y_high += dy;
    y_low += dy;
  }
  ctx->set_dirty();
}

int GrowConPoint::event_handler(glow_eEvent event, int x, int y, double fx, double fy)
{
  int sts;
  GlowPoint p1(p);
  glow_sPoint ptmp = trf * p;
  p1.x = ptmp.x;
  p1.y = ptmp.y;
  p1.zoom();

  if (event == ctx->event_move_node) {
    sts = arc.event_handler((void*)&p1, event, x, y, NULL);
    if (sts) {
      /* Register node for potential movement */
      ctx->move_insert(this);
    }
    return sts;
  }
  if (event == glow_eEvent_CursorMotion) {
    if (ctx->hot_found)
      sts = 0;
    else {
      sts = arc.event_handler((void*)&p1, event, x, y, NULL);
      if (sts)
        ctx->hot_found = 1;
    }
    if (sts && !hot
        && !(ctx->node_movement_active || ctx->node_movement_paste_active)) {
      ctx->gdraw->set_cursor(ctx->mw.window, glow_eDrawCursor_CrossHair);
      set_hot(1);
    }
    if (!sts && hot) {
      ctx->gdraw->set_cursor(ctx->mw.window, glow_eDrawCursor_Normal);
      set_hot(0);
    }
  } else {
    sts = arc.event_handler((void*)&p1, event, x, y, NULL);
  }
  if (sts)
    ctx->register_callback_object(glow_eObjectType_Node, this);
  return sts;
}

void GrowConPoint::save(std::ofstream& fp, glow_eSaveMode mode)
{
  if (mode == glow_eSaveMode_SubGraph)
    GlowConPoint::save(fp, mode);
  else {
    fp << int(glow_eSave_GrowConPoint) << '\n';
    fp << int(glow_eSave_GrowConPoint_n_name) << FSPACE << n_name << '\n';
    fp << int(glow_eSave_GrowConPoint_x_right) << FSPACE << x_right << '\n';
    fp << int(glow_eSave_GrowConPoint_x_left) << FSPACE << x_left << '\n';
    fp << int(glow_eSave_GrowConPoint_y_high) << FSPACE << y_high << '\n';
    fp << int(glow_eSave_GrowConPoint_y_low) << FSPACE << y_low << '\n';
    fp << int(glow_eSave_GrowConPoint_arc) << '\n';
    arc.save(fp, mode);
    fp << int(glow_eSave_GrowConPoint_conpoint_part) << '\n';
    GlowConPoint::save(fp, mode);
    fp << int(glow_eSave_End) << '\n';
  }
}

void GrowConPoint::open(std::ifstream& fp)
{
  int type = 0;
  int end_found = 0;
  char dummy[40];

  for (;;) {
    if (!fp.good()) {
      fp.clear();
      fp.getline(dummy, sizeof(dummy));
      printf("** Read error GrowConPoint: \"%d %s\"\n", type, dummy);
    }

    fp >> type;
    switch (type) {
    case glow_eSave_GrowConPoint:
      break;
    case glow_eSave_GrowConPoint_n_name:
      fp.get();
      fp.getline(n_name, sizeof(n_name));
      break;
    case glow_eSave_GrowConPoint_x_right:
      fp >> x_right;
      break;
    case glow_eSave_GrowConPoint_x_left:
      fp >> x_left;
      break;
    case glow_eSave_GrowConPoint_y_high:
      fp >> y_high;
      break;
    case glow_eSave_GrowConPoint_y_low:
      fp >> y_low;
      break;
    case glow_eSave_GrowConPoint_arc:
      arc.open(fp);
      break;
    case glow_eSave_GrowConPoint_conpoint_part:
      GlowConPoint::open(fp);
      break;
    case glow_eSave_End:
      end_found = 1;
      break;
    default:
      std::cout << "GrowConPoint:open syntax error\n";
      fp.getline(dummy, sizeof(dummy));
    }
    if (end_found)
      break;
  }
}

void GrowConPoint::draw(GlowWind* w, int ll_x, int ll_y, int ur_x, int ur_y)
{
  int tmp;

  if (ll_x > ur_x) {
    /* Shift */
    tmp = ll_x;
    ll_x = ur_x;
    ur_x = tmp;
  }
  if (ll_y > ur_y) {
    /* Shift */
    tmp = ll_y;
    ll_y = ur_y;
    ur_y = tmp;
  }

  if (x_right * w->zoom_factor_x - w->offset_x >= ll_x
      && x_left * w->zoom_factor_x - w->offset_x <= ur_x
      && y_high * w->zoom_factor_y - w->offset_y >= ll_y
      && y_low * w->zoom_factor_y - w->offset_y <= ur_y) {
    draw(w, (GlowTransform*)NULL, highlight, hot, NULL, NULL);
  }
}

void GrowConPoint::draw(GlowWind* w, int* ll_x, int* ll_y, int* ur_x, int* ur_y)
{
  int tmp;
  int obj_ur_x = int(x_right * w->zoom_factor_x) - w->offset_x;
  int obj_ll_x = int(x_left * w->zoom_factor_x) - w->offset_x;
  int obj_ur_y = int(y_high * w->zoom_factor_y) - w->offset_y;
  int obj_ll_y = int(y_low * w->zoom_factor_y) - w->offset_y;

  if (*ll_x > *ur_x) {
    /* Shift */
    tmp = *ll_x;
    *ll_x = *ur_x;
    *ur_x = tmp;
  }
  if (*ll_y > *ur_y) {
    /* Shift */
    tmp = *ll_y;
    *ll_y = *ur_y;
    *ur_y = tmp;
  }

  if (obj_ur_x >= *ll_x && obj_ll_x <= *ur_x && obj_ur_y >= *ll_y
      && obj_ll_y <= *ur_y) {
    draw(w, (GlowTransform*)NULL, highlight, hot, NULL, NULL);

    // Increase the redraw area
    if (obj_ur_x > *ur_x)
      *ur_x = obj_ur_x;
    if (obj_ur_y > *ur_y)
      *ur_y = obj_ur_y;
    if (obj_ll_x < *ll_x)
      *ll_x = obj_ll_x;
    if (obj_ll_y < *ll_y)
      *ll_y = obj_ll_y;
  }
}

void GrowConPoint::set_highlight(int on)
{
  if (highlight != on) {
    highlight = on;
    ctx->set_dirty();
  }
}

void GrowConPoint::select_region_insert(double ll_x, double ll_y, double ur_x,
    double ur_y, glow_eSelectPolicy select_policy)
{
  if (select_policy == glow_eSelectPolicy_Surround) {
    if (x_left > ll_x && x_right < ur_x && y_high < ur_y && y_low > ll_y)
      ctx->select_insert(this);
  } else {
    if (x_right > ll_x && x_left < ur_x && y_low < ur_y && y_high > ll_y)
      ctx->select_insert(this);
  }
}

void GrowConPoint::zoom()
{
  GlowConPoint::zoom();
  arc.zoom();
}

void GrowConPoint::nav_zoom()
{
  GlowConPoint::nav_zoom();
  arc.nav_zoom();
}

void GrowConPoint::draw(GlowWind* w, GlowTransform* t, int highlight, int hot,
    void* node, void* colornode)
{
  if (w == &ctx->navw) {
    return;
  }

  Matrix tmp = t ? (*t * trf) : trf;
  glow_sPoint ptmp = tmp * p;

  GlowPoint p1(p);
  p1.x = ptmp.x;
  p1.y = ptmp.y;
  p1.zoom();
  arc.draw(w, (void*)&p1, highlight, hot, NULL);
}

void GrowConPoint::erase(GlowWind* w, GlowTransform* t, int hot, void* node)
{
  hot = (w == &ctx->navw) ? 0 : hot;

  Matrix tmp = t ? (*t * trf) : trf;
  glow_sPoint ptmp = tmp * p;

  GlowPoint p1(p);
  p1.x = ptmp.x;
  p1.y = ptmp.y;
  p1.zoom();
  arc.erase(w, (void*)&p1, hot, NULL);
}

void GrowConPoint::set_transform(GlowTransform* t)
{
  trf.set(*t * trf);
  get_node_borders();
}

void GrowConPoint::align(double x, double y, glow_eAlignDirection direction)
{
  double dx, dy;

  switch (direction) {
  case glow_eAlignDirection_CenterVert:
    dx = x - (x_right + x_left) / 2;
    dy = 0;
    break;
  case glow_eAlignDirection_CenterHoriz:
    dx = 0;
    dy = y - (y_high + y_low) / 2;
    break;
  case glow_eAlignDirection_CenterCenter:
    dx = x - (x_right + x_left) / 2;
    dy = y - (y_high + y_low) / 2;
    break;
  case glow_eAlignDirection_Right:
    dx = x - x_right;
    dy = 0;
    break;
  case glow_eAlignDirection_Left:
    dx = x - x_left;
    dy = 0;
    break;
  case glow_eAlignDirection_Up:
    dx = 0;
    dy = y - y_high;
    break;
  case glow_eAlignDirection_Down:
    dx = 0;
    dy = y - y_low;
    break;
  }
  trf.move(dx, dy);
  x_right += dx;
  x_left += dx;
  y_high += dy;
  y_low += dy;
  ctx->set_dirty();
}

void GrowConPoint::get_node_borders(GlowTransform* t, double* x_right,
    double* x_left, double* y_high, double* y_low)
{
  Matrix tmp = t ? (*t * trf) : trf;
  glow_sPoint p = tmp * this->p;
  arc.get_borders(p.x, p.y, x_right, x_left, y_high, y_low, NULL);
}

void GrowConPoint::get_borders(GlowTransform* t, double* x_right,
    double* x_left, double* y_high, double* y_low)
{
  Matrix tmp = t ? (*t * trf) : trf;
  glow_sPoint p = tmp * this->p;

  *x_left = MIN(*x_left, p.x);
  *x_right = MAX(*x_right, p.x);
  *y_low = MIN(*y_low, p.y);
  *y_high = MAX(*y_high, p.y);
}

void GrowConPoint::export_flow(GlowExportFlow* ef)
{
  ef->conpoint(this);
}
